#include "instr.hh"
#include "koala.hh"

// This monster of a switch statement decodes all CPU instructions. I've
// decided against an explicit jump table, as such implementation both
// increases the code size and introduced additional overhead to the inner
// decoder loop.

// WARNING: The below code currently does not simulate any slips.
// This should be fixed soon (it's fairly easy to simulate, too.)

int
Koala::decode(Instr instr)
{
    if (catch_nops > 0) {
	if (instr)
	    nop_count = 0;
	else if (++nop_count >= catch_nops) {
	    msg("%d NOP instructions executed in a row. Bailing out.",
		catch_nops);
	    Tcl_Exit(1);
	}
    }

    switch (opcode(instr)) {
    case SPECIAL:
    {
	// Special instructions, decoded using the (function) field.
	switch (function(instr)) {
	case SLL:
	{
	    // Shift Left Logical ////////////////////////////////////////////
	    if (rd(instr)) {
		// This is optimized for gpr[0], as it comonly used as a NOP.
		if (trace_level >= print_instructions)
		    log("[%lx]\tsll\t%s, %s, %d", pc,
			regname[rd(instr)],
			regname[rt(instr)],
			shamt(instr));
		UInt32 x = gpr[rt(instr)];
		int s = shamt(instr);
		gpr[rd(instr)] = sign_extend<UInt64>(x << s, 32);
	    }
	    else {
		if (trace_level >= print_instructions)
		    log("[%lx]\tnop", pc);
	    }
	    return nothing_special;
	}
	case SRL:
	{
	    // Shift Right Logical ///////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsrl\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr));
	    UInt32 x = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    int s = shamt(instr);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 32);
	    return nothing_special;
	}
	case SRA:
	{
	    // Shift Right Arithmetic ////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsra\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr));
	    UInt32 x = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    int s = shamt(instr);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 32 - s);
	    return nothing_special;
	}
	case SLLV:
	{
	    // Shift Left Logical Variable ///////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsllv\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    regname[rs(instr)]);
	    UInt32 x = gpr[rt(instr)];
	    int s = bits(gpr[rs(instr)], 4, 0);
	    gpr[rd(instr)] = sign_extend<UInt64>(x << s, 32);
	    return nothing_special;
	}
	case SRLV:
	{
	    // Shift Right Logical Variable //////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsrlv\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    regname[rs(instr)]);
	    UInt32 x = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    int s = bits(gpr[rs(instr)], 4, 0);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 32);
	    return nothing_special;
	}
	case SRAV:
	{
	    // Shift Right Arithmetic Variable ///////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsrav\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    regname[rs(instr)]);
	    UInt32 x = gpr[rt(instr)];
	    int s = bits(gpr[rs(instr)], 4, 0);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 32 - s);
	    return nothing_special;
	}
	case JR:
	{
	    // Jump Register /////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tjr\t%s", pc,
		    regname[rs(instr)]);
	    branch_target = gpr[rs(instr)];
	    return bits(branch_target, 1, 0) ? instr_addr_error :
		branch_delay;
	}
	case JALR:
	{
	    // Jump And Link Register ////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tjalr\t%s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)]);
	    branch_target = gpr[rs(instr)];
	    gpr[31] = sign_extend<UInt64>(pc + 8, 64);
	    return bits(branch_target, 1, 0) ? instr_addr_error :
		branch_delay;
	}
	case SYSCALL:
	{
	    // System Call ///////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsyscall", pc);
	    process_syscall();
	    return nothing_special;
	}
	case BREAK:
	{
	    // Breakpoint ////////////////////////////////////////////////////
	    if (trace_level >= print_instructions || bits(instr, 31, 16) > 0)
		log("[%lx]\tbreak\t%d", pc, bits(instr, 31, 16));
#if 0
	    // Enable tracing level.
	    if (trace_level >= 0) {
		trace_level = bits(instr, 31, 16);
		if (trace_level == 666) {
		    msg("BREAK 666: bailing out");
		    Tcl_Exit(1);
		}
		msg("Trace level set to %d", trace_level);
	    }
	    else {
		process_breakpoint();
	    }
#endif
	    process_breakpoint();
	    return nothing_special;
	}
	case SYNC:
	{
	    // Synchronize ///////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsync", pc);
	    sync_bit = true;
	    return nothing_special;
	}
	case MFHI:
	{
	    // Move From HI //////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tmfhi\t%s", pc,
		    regname[rd(instr)]);
	    gpr[rd(instr)] = hi;
	    return nothing_special;
	}
	case MTHI:
	{
	    // Move To HI ////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tmthi\t%s", pc,
		    regname[rs(instr)]);
	    hi = gpr[rs(instr)];
	    return nothing_special;
	}
	case MFLO:
	{
	    // Move From LO //////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tmflo\t%s", pc,
		    regname[rd(instr)]);
	    gpr[rd(instr)] = lo;
	    return nothing_special;
	}
	case MTLO:
	{
	    // Move To LO ////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tmtlo\t%s", pc,
		    regname[rs(instr)]);
	    lo = gpr[rs(instr)];
	    return nothing_special;
	}
	case DSLLV:
	{
	    // Doubleword Shift Left Logical Variable ////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsllv\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    regname[rs(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rt(instr)];
	    int s = bits(gpr[rs(instr)], 5, 0);
	    gpr[rd(instr)] = sign_extend<UInt64>(x << s, 64);
	    return nothing_special;
	}
	case DSRLV:
	{
	    // Doubleword Shift Right Logical Variable ///////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsrlv\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    regname[rs(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    int s = bits(gpr[rs(instr)], 5, 0);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 64);
	    return nothing_special;
	}
	case DSRAV:
	{
	    // Doubleword Shift Right Arithmetic Variable ////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsrav\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    regname[rs(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rt(instr)];
	    int s = bits(gpr[rs(instr)], 5, 0);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 64 - s);
	    return nothing_special;
	}
	case MULT:
	{
	    // Multiply //////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tmult\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    Int32 x = sign_extend<Int32>(gpr[rs(instr)], 32);
	    Int32 y = sign_extend<Int32>(gpr[rt(instr)], 32);
	    MulResult<Int32> r = multiply(x, y);
	    lo = sign_extend<UInt64>(r.lo, 32);
	    hi = sign_extend<UInt64>(r.hi, 32);
	    return nothing_special;
	}
	case MULTU:
	{
	    // Multiply Unsigned /////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tmultu\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    UInt32 x = zero_extend<UInt32>(gpr[rs(instr)], 32);
	    UInt32 y = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    MulResult<UInt32> r = multiply(x, y);
	    lo = sign_extend<UInt64>(r.lo, 32);
	    hi = sign_extend<UInt64>(r.hi, 32);
	    return nothing_special;
	}
	case DIV:
	{
	    // Divide ////////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdiv\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    Int32 y = sign_extend<Int32>(gpr[rt(instr)], 32);
	    if (y) {
		Int32 x = sign_extend<Int32>(gpr[rs(instr)], 32);
		DivResult<Int32> r = divide(x, y);
		lo = sign_extend<UInt64>(r.quot, 32);
		hi = sign_extend<UInt64>(r.rem, 32);
	    }
	    return nothing_special;
	}
	case DIVU:
	{
	    // Divide Unsigned ///////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdivu\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    UInt32 y = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    if (y) {
		UInt32 x = zero_extend<UInt32>(gpr[rs(instr)], 32);
		DivResult<UInt32> r = divide(x, y);
		lo = sign_extend<UInt64>(r.quot, 32);
		hi = sign_extend<UInt64>(r.rem, 32);
	    }
	    return nothing_special;
	}
	case DMULT:
	{
	    // Doubleword Multiply ///////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdmult\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    Int64 x = gpr[rs(instr)];
	    Int64 y = gpr[rt(instr)];
	    MulResult<Int64> r = multiply(x, y);
	    lo = sign_extend<UInt64>(r.lo, 64);
	    hi = sign_extend<UInt64>(r.hi, 64);
	    return nothing_special;
	}
	case DMULTU:
	{
	    // Doubleword Multiply Unsigned //////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdmultu\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = zero_extend<UInt64>(gpr[rs(instr)], 64);
	    UInt64 y = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    MulResult<UInt64> r = multiply(x, y);
	    lo = sign_extend<UInt64>(r.lo, 64);
	    hi = sign_extend<UInt64>(r.hi, 64);
	    return nothing_special;
	}
	case DDIV:
	{
	    // Doubleword Divide /////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tddiv\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    Int64 y = gpr[rt(instr)];
	    if (y) {
		Int64 x = gpr[rs(instr)];
		DivResult<Int64> r = divide(x, y);
		lo = sign_extend<UInt64>(r.quot, 64);
		hi = sign_extend<UInt64>(r.rem, 64);
	    }
	    return nothing_special;
	}
	case DDIVU:
	{
	    // Doubleword Divide Unsigned ////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tddivu\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 y = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    if (y) {
		UInt64 x = zero_extend<UInt64>(gpr[rs(instr)], 64);
		DivResult<UInt64> r = divide(x, y);
		lo = sign_extend<UInt64>(r.quot, 64);
		hi = sign_extend<UInt64>(r.rem, 64);
	    }
	    return nothing_special;
	}
	case ADD:
	    {
	    // Add ///////////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tadd\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    UInt32 x = gpr[rs(instr)];
	    UInt32 y = gpr[rt(instr)];
	    UInt32 z = x + y;
	    // Overflow occurs is sign(x) == sign(y) != sign(z).
	    if (bit(x ^ y, 31) == 0 && bit(x ^ z, 31) != 0)
		process_integer_overflow();
	    gpr[rd(instr)] = sign_extend<UInt64>(z, 32);
	    return nothing_special;
	}
	case ADDU:
	{
	    // Add Unsigned //////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\taddu\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    UInt32 x = gpr[rs(instr)];
	    UInt32 y = gpr[rt(instr)];
	    gpr[rd(instr)] = sign_extend<UInt64>(x + y, 32);
	    return nothing_special;
	}
	case SUB:
	{
	    // Subtract //////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsub\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    Int32 x = zero_extend<UInt32>(gpr[rs(instr)], 32);
	    Int32 y = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    Int32 z = UInt32(x) - UInt32(y);
	    if ((y < 0 && z < x) || (y > 0 && z > x))
		process_integer_overflow();
	    gpr[rd(instr)] = sign_extend<UInt64>(z, 32);
	    return nothing_special;
	}
	case SUBU:
	{
	    // Subtract Unsigned /////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsubu\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    UInt32 x = zero_extend<UInt32>(gpr[rs(instr)], 32);
	    UInt32 y = zero_extend<UInt32>(gpr[rt(instr)], 32);
	    gpr[rd(instr)] = sign_extend<UInt64>(x - y, 32);
	    return nothing_special;
	}
	case AND:
	{
	    // And ///////////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tand\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    gpr[rd(instr)] = gpr[rs(instr)] & gpr[rt(instr)];
	    return nothing_special;
	}
	case OR:
	{
	    // Or ////////////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tor\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    gpr[rd(instr)] = gpr[rs(instr)] | gpr[rt(instr)];
	    return nothing_special;
	}
	case XOR:
	{
	    // Exclusive Or //////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\txor\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    gpr[rd(instr)] = gpr[rs(instr)] ^ gpr[rt(instr)];
	    return nothing_special;
	}
	case NOR:
	{
	    // Nor ///////////////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tnor\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    gpr[rd(instr)] = ~(gpr[rs(instr)] | gpr[rt(instr)]);
	    return nothing_special;
	}
	case SLT:
	{
	    // Set On Less Than //////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tslt\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    Int64 x = gpr[rs(instr)];
	    Int64 y = gpr[rt(instr)];
	    gpr[rd(instr)] = (x < y);
	    return nothing_special;
	}
	case SLTU:
	{
	    // Set On Less Than Unsigned /////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tsltu\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    gpr[rd(instr)] = (gpr[rs(instr)] < gpr[rt(instr)]);
	    return nothing_special;
	}
	case DADD:
	{
	    // Doubleword Add ////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdadd\t%s, %s, %s", pc, regname[rd(instr)],
		    regname[rs(instr)], regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rs(instr)];
	    UInt64 y = gpr[rt(instr)];
	    UInt64 z = x + y;
	    // Overflow occurs is sign(x) == sign(y) != sign(z).
	    if (bit(x ^ y, 63) == 0 && bit(x ^ z, 63) != 0)
		process_integer_overflow();
	    gpr[rd(instr)] = sign_extend<UInt64>(z, 64);
	    return nothing_special;
	}
	case DADDU:
	{
	    // Doubleword Add Unsigned ///////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdaddu\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rs(instr)];
	    UInt64 y = gpr[rt(instr)];
	    gpr[rd(instr)] = sign_extend<UInt64>(x + y, 64);
	    return nothing_special;
	}
	case DSUB:
	{
	    // Doubleword Subtract ///////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsub\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    Int64 x = zero_extend<UInt64>(gpr[rs(instr)], 64);
	    Int64 y = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    Int64 z = UInt64(x) - UInt64(y);
	    if ((y < 0 && z < x) || (y > 0 && z > x))
		process_integer_overflow();
	    gpr[rd(instr)] = sign_extend<UInt64>(z, 64);
	    return nothing_special;
	}
	case DSUBU:
	{
	    // Doubleword Subtract Unsigned //////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsubu\t%s, %s, %s", pc,
		    regname[rd(instr)],
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = zero_extend<UInt64>(gpr[rs(instr)], 64);
	    UInt64 y = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    gpr[rd(instr)] = sign_extend<UInt64>(x - y, 64);
	    return nothing_special;
	}
	case TGE:
	{
	    // Trap If Greater Than Or Equal /////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttge\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    Int64 x = gpr[rs(instr)];
	    Int64 y = gpr[rt(instr)];
	    if (x >= y)
		process_trap();
	    return nothing_special;
	}
	case TGEU:
	{
	    // Trap If Greater Than Or Equal Unsigned ////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttgeu\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (gpr[rs(instr)] >= gpr[rt(instr)])
		process_trap();
	    return nothing_special;
	}
	case TLT:
	{
	    // Trap If Less Than /////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttlt\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    Int64 x = gpr[rs(instr)];
	    Int64 y = gpr[rt(instr)];
	    if (x < y)
		process_trap();
	    return nothing_special;
	}
	case TLTU:
	{
	    // Trap If Less Than Unsigned ////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttltu\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (gpr[rs(instr)] < gpr[rt(instr)])
		process_trap();
	    return nothing_special;
	}
	case TEQ:
	{
	    // Trap If Equal /////////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tteq\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (gpr[rs(instr)] == gpr[rt(instr)])
		process_trap();
	    return nothing_special;
	}
	case TNE:
	{
	    // Trap If Not Equal /////////////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttne\t%s, %s", pc,
		    regname[rs(instr)],
		    regname[rt(instr)]);
	    if (gpr[rs(instr)] != gpr[rt(instr)])
		process_trap();
	    return nothing_special;
	}
	case DSLL:
	{
	    // Doubleword Shift Left Logical /////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsll\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr));
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rt(instr)];
	    int s = shamt(instr);
	    gpr[rd(instr)] = sign_extend<UInt64>(x << s, 64);
	    return nothing_special;
	}
	case DSRL:
	{
	    // Doubleword Shift Right Logical ////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsrl\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr));
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    int s = shamt(instr);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 64);
	    return nothing_special;
	}
	case DSRA:
	{
	    // Doubleword Shift Right Arithmetic /////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsra\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr));
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rt(instr)];
	    int s = shamt(instr);
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 64 - s);
	    return nothing_special;
	}
	case DSLL32:
	{
	    // Doubleword Shift Left Logical + 32 ////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsll\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr) + 32);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rt(instr)];
	    int s = shamt(instr) + 32;
	    gpr[rd(instr)] = sign_extend<UInt64>(x << s, 64);
	    return nothing_special;
	}
	case DSRL32:
	{
	    // Doubleword Shift Right Logical + 32 ///////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsrl\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr) + 32);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = zero_extend<UInt64>(gpr[rt(instr)], 64);
	    int s = shamt(instr) + 32;
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 64);
	    return nothing_special;
	}
	case DSRA32:
	{
	    // Doubleword Shift Right Arithmetic + 32 ////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tdsra\t%s, %s, %d", pc,
		    regname[rd(instr)],
		    regname[rt(instr)],
		    shamt(instr) + 32);
	    if (!allow_xinstr())
		process_reserved_instruction();
	    UInt64 x = gpr[rt(instr)];
	    int s = shamt(instr) + 32;
	    gpr[rd(instr)] = sign_extend<UInt64>(x >> s, 64 - s);
	    return nothing_special;
	}
	default:
	    // Reserved instruction.
	    if (trace_level >= print_instructions)
		log("[%lx]\t.long\t0x%08x", pc, instr);
	    process_reserved_instruction();
	    return nothing_special;
	}
    }
    case REGIMM:
    {
	// Conditional branches, decoded using the (rt) field.
	switch (rt(instr)) {
	case BLTZ:
	{
	    // Branch On Less Than Zero //////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbltz\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	    Int64 x = gpr[rs(instr)];
	    if (x < 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
	    }
	    else
		branch_target = pc + 8;
	    return branch_delay;
	}
	case BGEZ:
	{
	    // Branch On Greater Than Or Equal To Zero ///////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbgez\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	    Int64 x = gpr[rs(instr)];
	    if (x >= 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
	    }
	    else
		branch_target = pc + 8;
	    return branch_delay;
	}
	case BLTZL:
	{
	    // Branch On Less Than Zero Likely ///////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbltzl\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	    Int64 x = gpr[rs(instr)];
	    if (x < 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
		return branch_delay;
	    }
	    else {
		// Nullify (skip) the next instruction.
		pc += 4;
		return nothing_special;
	    }
	}
	case BGEZL:
	{
	    // Branch On Greater Than Or Equal To Zero Likely ////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbgezl\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	    Int64 x = gpr[rs(instr)];
	    if (x >= 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
		return branch_delay;
	    }
	    else {
		// Nullify (skip) the next instruction.
		pc += 4;
		return nothing_special;
	    }
	}
	case TGEI:
	{
	    // Trap If Greater Than Or Equal Immediate ///////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttgei\t%s, 0x%04x", pc,
		    regname[rs(instr)],
		    zero_extend<UInt32>(immediate(instr), 16));
	    Int64 x = gpr[rs(instr)];
	    Int64 y = sign_extend<Int64>(immediate(instr), 16);
	    if (x >= y)
		process_trap();
	    return nothing_special;
	}
	case TGEIU:
	{
	    // Trap If Greater Than Or Equal Immediate Unsigned //////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttgeiu\t%s, 0x%04x", pc,
		    regname[rs(instr)],
		    zero_extend<UInt32>(immediate(instr), 16));
	    UInt64 x = gpr[rs(instr)];
	    UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	    if (x >= y)
		process_trap();
	    return nothing_special;
	}
	case TLTI:
	{
	    // Trap If Less Than Immediate ///////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttlti\t%s, 0x%04x", pc,
		    regname[rs(instr)],
		    zero_extend<UInt32>(immediate(instr), 16));
	    Int64 x = gpr[rs(instr)];
	    Int64 y = sign_extend<Int64>(immediate(instr), 16);
	    if (x < y)
		process_trap();
	    return nothing_special;
	}
	case TLTIU:
	{
	    // Trap If Less Than Immediate Unsigned //////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttltiu\t%s, 0x%04x", pc,
		    regname[rs(instr)],
		    zero_extend<UInt32>(immediate(instr), 16));
	    UInt64 x = gpr[rs(instr)];
	    UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	    if (x < y)
		process_trap();
	    return nothing_special;
	}
	case TEQI:
	{
	    // Trap If Equal Immediate ///////////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tteqi\t%s, 0x%04x", pc,
		    regname[rs(instr)],
		    zero_extend<UInt32>(immediate(instr), 16));
	    UInt64 x = gpr[rs(instr)];
	    UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	    if (x == y)
		process_trap();
	    return nothing_special;
	}
	case TNEI:
	{
	    // Trap If Not Equal Immediate ///////////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\ttnei\t%s, 0x%04x", pc,
		    regname[rs(instr)],
		    zero_extend<UInt32>(immediate(instr), 16));
	    UInt64 x = gpr[rs(instr)];
	    UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	    if (x != y)
		process_trap();
	    return nothing_special;
	}
	case BLTZAL:
	{
	    // Branch On Less Than Zero And Link /////////////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbltzal\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + sign_extend<VA>(offset(instr), 16) << 2);
	    Int64 x = gpr[rs(instr)];
	    gpr[31] = sign_extend<UInt64>(pc + 8, 64);
	    if (x < 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
	    }
	    else
		branch_target = pc + 8;
	    return branch_delay;
	}
	case BGEZAL:
	{
	    // Branch On Greater Than Or Equal To Zero And Link //////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbltzal\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + sign_extend<VA>(offset(instr), 16) << 2);
	    Int64 x = gpr[rs(instr)];
	    gpr[31] = sign_extend<UInt64>(pc + 8, 64);
	    if (x >= 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
	    }
	    else
		branch_target = pc + 8;
	    return branch_delay;
	}
	case BLTZALL:
	{
	    // Branch On Less Than Zero And Link Likely //////////////////////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbltzall\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	    Int64 x = gpr[rs(instr)];
	    gpr[31] = sign_extend<UInt64>(pc + 8, 64);
	    if (x < 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
		return branch_delay;
	    }
	    else {
		// Nullify (skip) the next instruction.
		pc += 4;
		return nothing_special;
	    }
	}
	case BGEZALL:
	{
	    // Branch On Greater Than Or Equal To Zero And Link Likely ///////
	    if (trace_level >= print_instructions)
		log("[%lx]\tbgezall\t%s, 0x%016lx", pc,
		    regname[rs(instr)],
		    pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	    Int64 x = gpr[rs(instr)];
	    gpr[31] = sign_extend<UInt64>(pc + 8, 64);
	    if (x >= 0) {
		VA off = sign_extend<VA>(offset(instr), 16);
		branch_target = pc + 4 + (off << 2);
		return branch_delay;
	    }
	    else {
		// Nullify (skip) the next instruction.
		pc += 4;
		return nothing_special;
	    }
	}
	default:
	    // Reserved instruction.
	    if (trace_level >= print_instructions)
		log("[%lx]\t.long\t0x%08x", pc, instr);
	    process_reserved_instruction();
	    return nothing_special;
	}
    }
    case J:
    {
	// Jump //////////////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tj\t0x%016lx", pc,
		clear_bits(pc + 4, 27, 0) | (target(instr) << 2));
	VA msb = clear_bits(pc + 4, 27, 0);
	branch_target = msb | (target(instr) << 2);
	return branch_delay;
    }
    case JAL:
    {
	// Jump And Link /////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tjal\t0x%016lx", pc,
		clear_bits(pc + 4, 27, 0) | (target(instr) << 2));
	gpr[31] = sign_extend<UInt64>(pc + 8, 64);
	VA msb = clear_bits(pc + 4, 27, 0);
	branch_target = msb | (target(instr) << 2);
	return branch_delay;
    }
    case BEQ:
    {
	// Branch On Equal ///////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tbeq\t%s, %s, 0x%016lx", pc,
		regname[rs(instr)],
		regname[rt(instr)],
		pc + 4 + (sign_extend<Int64>(offset(instr), 16) << 2));
	if (gpr[rs(instr)] == gpr[rt(instr)]) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	}
	else
	    branch_target = pc + 8;
	return branch_delay;
    }
    case BNE:
    {
	// Branch On Not Equal ///////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tbne\t%s, %s, 0x%016lx", pc,
		regname[rs(instr)],
		regname[rt(instr)],
		pc + 4 + (sign_extend<Int64>(offset(instr), 16) << 2));
	if (gpr[rs(instr)] != gpr[rt(instr)]) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	}
	else
	    branch_target = pc + 8;
	return branch_delay;
    }
    case BLEZ:
    {
	// Branch On Less Than Or Equal To Zero //////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tblez\t%s, 0x%016lx", pc,
		regname[rs(instr)],
		pc + 4 + (sign_extend<Int64>(offset(instr), 16) << 2));
	Int64 x = gpr[rs(instr)];
	if (x <= 0) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	}
	else
	    branch_target = pc + 8;
	return branch_delay;
    }
    case BGTZ:
    {
	// Branch On Greater Than Zero ///////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tbgtz\t%s, 0x%016lx", pc,
		regname[rs(instr)],
		pc + 4 + (sign_extend<Int64>(offset(instr), 16) << 2));
	Int64 x = gpr[rs(instr)];
	if (x > 0) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	}
	else
	    branch_target = pc + 8;
	return branch_delay;
    }
    case ADDI:
    {
	// Add Immediate /////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\taddi\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	UInt32 x = gpr[rs(instr)];
	UInt32 y = sign_extend<UInt32>(immediate(instr), 16);
	UInt32 z = x + y;
	// Overflow occurs is sign(x) == sign(y) != sign(z).
	if (bit(x ^ y, 31) == 0 && bit(x ^ z, 31) != 0)
	    process_integer_overflow();
	gpr[rt(instr)] = sign_extend<UInt64>(z, 32);
	return nothing_special;
    }
    case ADDIU:
    {
	// Add Immediate Unsigned ////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\taddiu\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	UInt32 x = gpr[rs(instr)];
	UInt32 y = sign_extend<UInt32>(immediate(instr), 16);
	gpr[rt(instr)] = sign_extend<UInt64>(x + y, 32);
	return nothing_special;
    }
    case SLTI:
    {
	// Set On Less Than Immediate ////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tslti\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	Int64 x = gpr[rs(instr)];
	Int64 y = sign_extend<Int64>(immediate(instr), 16);
	gpr[rt(instr)] = (x < y);
	return nothing_special;
    }
    case SLTIU:
    {
	// Set On Less Than Immediate Unsigned ///////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsltiu\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	UInt64 x = gpr[rs(instr)];
	UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	gpr[rt(instr)] = (x < y);
	return nothing_special;
    }
    case ANDI:
    {
	// And Immediate /////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tandi\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	UInt16 x = gpr[rs(instr)];
	UInt16 imm = zero_extend<UInt64>(immediate(instr), 16);
	gpr[rt(instr)] = zero_extend<UInt64>(x & imm, 16);
	return nothing_special;
    }
    case ORI:
    {
	// Or Immediate //////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tori\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	UInt64 x = gpr[rs(instr)];
	UInt64 imm = zero_extend<UInt64>(immediate(instr), 16);
	gpr[rt(instr)] = x | imm;
	return nothing_special;
    }
    case XORI:
    {
	// Exclusive Or Immediate ////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\txori\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<UInt32>(immediate(instr), 16));
	UInt64 x = gpr[rs(instr)];
	UInt64 imm = zero_extend<UInt64>(immediate(instr), 16);
	gpr[rt(instr)] = x ^ imm;
	return nothing_special;
    }
    case LUI:
    {
	// Load Upper Immediate //////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlui\t%s, 0x%04x", pc,
		regname[rt(instr)],
		zero_extend<UInt32>(immediate(instr), 32));
	UInt32 imm = immediate(instr);
	gpr[rt(instr)] = sign_extend<UInt64>(imm << 16, 32);
	return nothing_special;
    }
    case COP0:
    {
	// Coprocessor 0 Operation ///////////////////////////////////////////
	if (trace_level >= print_instructions) {
	    if (bit(instr, 25)) {
		switch (funct(instr)) {
		case TLBR:
		    log("[%lx]\ttlbr", pc);
		    break;
		case TLBWI:
		    log("[%lx]\ttlbwi", pc);
		    break;
		case TLBWR:
		    log("[%lx]\ttlbwr", pc);
		    break;
		case TLBP:
		    log("[%lx]\ttlbp", pc);
		    break;
		case ERET:
		    log("[%lx]\teret", pc);
		    break;
		case WAIT:
		    log("[%lx]\twait", pc);
		    break;
		default:
		    log("[%lx]\t.long\t0x%08x // cop0", pc, instr);
		}
	    }
	    else {
		switch (rs(instr)) {
		case MFCz:
		    log("[%lx]\tmfc0\t%s, %d", pc,
			regname[rt(instr)],
			rd(instr));
		    break;
		case DMFCz:
		    log("[%lx]\tdmfc0\t%s, %d", pc,
			regname[rt(instr)],
			rd(instr));
		    break;
		case CFCz:
		    log("[%lx]\t.long\t0x%08x // cfc0", pc, instr);
		    break;
		case MTCz:
		    log("[%lx]\tmtc0\t%s, %d", pc,
			regname[rt(instr)],
			rd(instr));
		    break;
		case DMTCz:
		    log("[%lx]\tdmtc0\t%s, %d", pc,
			regname[rt(instr)],
			rd(instr));
		    break;
		case CTCz:
		    log("[%lx]\t.long\t0x%08x // ctc0", pc, instr);
		    break;
		case BCz:
		    log("[%lx]\t.long\t0x%08x // bc0", pc, instr);
		    break;
		default:
		    log("[%lx]\t.long\t0x%08x // cop0", pc, instr);
		}
	    }
	}
	return decode_cop0(instr);
    }
    case COP1:
    {
	// Coprocessor 1 Operation ///////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // cop1", pc, instr);
	return decode_cop1(instr);
    }
    case COP2:
    {
	// Coprocessor 2 Operation ///////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // cop2", pc, instr);
	process_reserved_instruction();
	return nothing_special;
    }
    case BEQL:
    {
	// Branch On Equal Likely ////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tbeql\t%s, %s, 0x%016lx", pc,
		regname[rs(instr)],
		regname[rt(instr)],
		pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	if (gpr[rs(instr)] == gpr[rt(instr)]) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	    return branch_delay;
	}
	else {
	    // Nullify (skip) the next instruction.
	    pc += 4;
	    return nothing_special;
	}
    }
    case BNEL:
    {
	// Branch On Not Equal Likely ////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tbnel\t%s, %s, 0x%016lx", pc,
		regname[rs(instr)],
		regname[rt(instr)],
		pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	if (gpr[rs(instr)] != gpr[rt(instr)]) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	    return branch_delay;
	}
	else {
	    // Nullify (skip) the next instruction.
	    pc += 4;
	    return nothing_special;
	}
    }
    case BLEZL:
    {
	// Branch On Less Than Or Equal To Zero Likely ///////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tblezl\t%s, 0x%016lx", pc,
		regname[rs(instr)],
		pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	Int64 x = gpr[rs(instr)];
	if (x <= 0) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	    return branch_delay;
	}
	else {
	    // Nullify (skip) the next instruction.
	    pc += 4;
	    return nothing_special;
	}
    }
    case BGTZL:
    {
	// Branch On Greater Than Zero Likely ////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tbgtzl\t%s, 0x%016lx", pc,
		regname[rs(instr)],
		pc + 4 + (sign_extend<VA>(offset(instr), 16) << 2));
	Int64 x = gpr[rs(instr)];
	if (x > 0) {
	    VA off = sign_extend<VA>(offset(instr), 16);
	    branch_target = pc + 4 + (off << 2);
	    return branch_delay;
	}
	else {
	    // Nullify (skip) the next instruction.
	    pc += 4;
	    return nothing_special;
	}
    }
    case DADDI:
    {
	// Doubleword Add Immediate //////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tdaddi\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		zero_extend<Int32>(immediate(instr), 16));
	if (!allow_xinstr())
	    process_reserved_instruction();
	UInt64 x = gpr[rs(instr)];
	UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	UInt64 z = x + y;
	// Overflow occurs is sign(x) == sign(y) != sign(z).
	if (bit(x ^ y, 63) == 0 && bit(x ^ z, 63) != 0)
	    process_integer_overflow();
	gpr[rt(instr)] = sign_extend<UInt64>(z, 64);
	return nothing_special;
    }
    case DADDIU:
    {
	// Doubleword Add Immediate Unsigned /////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tdaddiu\t%s, %s, 0x%04x", pc,
		regname[rt(instr)],
		regname[rs(instr)],
		sign_extend<Int32>(immediate(instr), 16));
	if (!allow_xinstr())
	    process_reserved_instruction();
	UInt64 x = gpr[rs(instr)];
	UInt64 y = sign_extend<UInt64>(immediate(instr), 16);
	gpr[rt(instr)] = sign_extend<UInt64>(x + y, 64);
	return nothing_special;
    }
    case LDL:
    {
	// Load Doubleword Left //////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tldl\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt64 mem = load<doubleword>(round_down(va, 8), round_down(pa, 8));
	UInt64 reg = gpr[rt(instr)];
	int syscmd = bits(va, 2, 0);
	if (!big_endian_cpu())
	    syscmd ^= bitmask(2, 0);
	reg = set_bits(reg, mem, 63, syscmd * 8);
	gpr[rt(instr)] = sign_extend<UInt64>(reg, 64);
	return nothing_special;
    }
    case LDR:
    {
	// Load Doubleword Right /////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tldr\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt64 mem = load<doubleword>(round_down(va, 8), round_down(pa, 8));
	UInt64 reg = gpr[rt(instr)];
	int syscmd = bits(va, 2, 0);
	if (big_endian_cpu())
	    syscmd ^= bitmask(2, 0);
	reg = set_bits(reg, bits(mem, 63, syscmd * 8), 63 - syscmd * 8, 0);
	gpr[rt(instr)] = sign_extend<UInt64>(reg, 64);
	return nothing_special;
    }
    case LB:
    {
	// Load Byte /////////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlb\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	PA pa = translate_vaddr(va, data_load);
	UInt8 x = load<byte>(va, pa);
	gpr[rt(instr)] = sign_extend<UInt64>(x, 8);
	return nothing_special;
    }
    case LH:
    {
	// Load Halfword /////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlh\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bit(va, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt16 x = load<halfword>(va, pa);
	gpr[rt(instr)] = sign_extend<UInt64>(x, 16);
	return nothing_special;
    }
    case LWL:
    {
	// Load Word Left ////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlwl\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt32 mem = load<word>(round_down(va, 4), round_down(pa, 4));
	UInt32 reg = gpr[rt(instr)];
	int syscmd = bits(va, 1, 0);
	if (!big_endian_cpu())
	    syscmd ^= bitmask(1, 0);
	reg = set_bits(reg, mem, 31, syscmd * 8);
	gpr[rt(instr)] = sign_extend<UInt64>(reg, 32);
	return nothing_special;
    }
    case LW:
    {
	// Load Word /////////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlw\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 1, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt32 x = load<word>(va, pa);
	gpr[rt(instr)] = sign_extend<UInt64>(x, 32);
	return nothing_special;
    }
    case LBU:
    {
	// Load Byte Unsigned ////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlbu\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	PA pa = translate_vaddr(va, data_load);
	UInt8 x = load<byte>(va, pa);
	gpr[rt(instr)] = zero_extend<UInt64>(x, 8);
	return nothing_special;
    }
    case LHU:
    {
	// Load Halfword Unsigned ////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlhu\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bit(va, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt16 x = load<halfword>(va, pa);
	gpr[rt(instr)] = zero_extend<UInt64>(x, 16);
	return nothing_special;
    }
    case LWR:
    {
	// Load Word Right ///////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlwr\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt32 mem = load<word>(round_down(va, 4), round_down(pa, 4));
	UInt32 reg = gpr[rt(instr)];
	int syscmd = bits(va, 1, 0);
	if (big_endian_cpu())
	    syscmd ^= bitmask(1, 0);
	reg = set_bits(reg, bits(mem, 31, syscmd * 8), 31 - syscmd * 8, 0);
	gpr[rt(instr)] = sign_extend<UInt64>(reg, 32);
	return nothing_special;
    }
    case LWU:
    {
	// Load Word Unsigned ////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlwu\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 1, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt32 x = load<word>(va, pa);
	gpr[rt(instr)] = zero_extend<UInt64>(x, 32);
	return nothing_special;
    }
    case SB:
    {
	// Store Byte ////////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsb\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	PA pa = translate_vaddr(va, data_store);
	store<byte>(gpr[rt(instr)], va, pa);
	return nothing_special;
    }
    case SH:
    {
	// Store Halfword ////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsh\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bit(va, 0))
	    process_address_error(data_store, va);
	PA pa = translate_vaddr(va, data_store);
	store<halfword>(gpr[rt(instr)], va, pa);
	return nothing_special;
    }
    case SWL:
    {
	// Store Word Left ///////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tswl\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt32 mem = load<word>(round_down(va, 4), round_down(pa, 4));
	UInt32 reg = gpr[rt(instr)];
	int syscmd = bits(va, 1, 0);
	if (!big_endian_cpu())
	    syscmd ^= bitmask(1, 0);
	mem = set_bits(mem, bits(reg, 31, syscmd*8), 31 - syscmd * 8, 0);
	store<word>(mem, round_down(va, 4), round_down(pa, 4));
	return nothing_special;
    }
    case SW:
    {
	// Store Word ////////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsw\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 1, 0))
	    process_address_error(data_store, va);
	PA pa = translate_vaddr(va, data_store);
	store<word>(gpr[rt(instr)], va, pa);
	return nothing_special;
    }
    case SDL:
    {
	// Store Doubleword Left /////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsdl\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt64 mem = load<doubleword>(round_down(va, 8), round_down(pa, 8));
	UInt64 reg = gpr[rt(instr)];
	int syscmd = bits(va, 2, 0);
	if (!big_endian_cpu())
	    syscmd ^= bitmask(2, 0);
	mem = set_bits(mem, bits(reg, 63, syscmd*8), 63 - syscmd * 8, 0);
	store<doubleword>(mem, round_down(va, 8), round_down(pa, 8));
	return nothing_special;
    }
    case SDR:
    {
	// Store Doubleword Right ////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsdr\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt64 mem = load<doubleword>(round_down(va, 8), round_down(pa, 8));
	UInt64 reg = gpr[rt(instr)];
	int syscmd = bits(va, 2, 0);
	if (big_endian_cpu())
	    syscmd ^= bitmask(2, 0);
	mem = set_bits(mem, reg, 63, syscmd * 8);
	store<doubleword>(mem, round_down(va, 8), round_down(pa, 8));
	return nothing_special;
    }
    case SWR:
    {
	// Store Word Right //////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tswr\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	// WARNING: see the comment in unaligned.cc.
	PA pa = translate_vaddr(va, data_load);
	UInt32 mem = load<word>(round_down(va, 4), round_down(pa, 4));
	UInt32 reg = gpr[rt(instr)];
	int syscmd = bits(va, 1, 0);
	if (big_endian_cpu())
	    syscmd ^= bitmask(1, 0);
	mem = set_bits(mem, reg, 31, syscmd * 8);
	store<word>(mem, round_down(va, 4), round_down(pa, 4));
	return nothing_special;
    }
    case CACHE:
    {
	// Cache /////////////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tcache\t%d, %d(%s)", pc,
		bits(instr, 20, 16),
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	return decode_cache(instr);
    }
    case LL:
    {
	// Load Linked ///////////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tll\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 1, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt32 x = load<word>(va, pa);
	gpr[rt(instr)] = sign_extend<UInt64>(x, 32);
	cp0[LLAddr] = bits(pa, 35, 4);
	sync_bit = true;
	ll_bit = true;
	return nothing_special;
    }
    case LWC1:
    {
	// Load Word to Coprocessor 1 ////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // lwc1", pc, instr);
	return decode_lwc1(instr);
    }
    case LWC2:
    {
	// Load Word to Coprocessor 2 ////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // lwc2", pc, instr);
	process_reserved_instruction();
	return nothing_special;
    }
    case LLD:
    {
	// Load Linked Doubleword ////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tlld\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 2, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt64 x = load<doubleword>(va, pa);
	gpr[rt(instr)] = sign_extend<UInt64>(x, 64);
	cp0[LLAddr] = bits(pa, 35, 4);
	sync_bit = true;
	ll_bit = true;
	return nothing_special;
    }
    case LDC1:
    {
	// Load Doubleword To Coprocessor 1 //////////////////////////////////
	return decode_ldc1(instr);
    }
    case LDC2:
    {
	// Load Doubleword To Coprocessor 2 //////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // ldc2", pc, instr);
	process_reserved_instruction();
	return nothing_special;
    }
    case LD:
    {
	// Load Doubleword ///////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tld\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 2, 0))
	    process_address_error(data_load, va);
	PA pa = translate_vaddr(va, data_load);
	UInt64 x = load<doubleword>(va, pa);
	gpr[rt(instr)] = sign_extend<UInt64>(x, 64);
	return nothing_special;
    }
    case SC:
    {
	// Store Conditional /////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsc\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (sync_bit)
	    sync();
	if (bits(va, 1, 0))
	    process_address_error(data_store, va);
	PA pa = translate_vaddr(va, data_store);
	if (ll_bit)
	    store<word>(gpr[rt(instr)], va, pa);
	gpr[rt(instr)] = ll_bit;
	sync_bit = true;
	ll_bit = false;
	return nothing_special;
    }
    case SWC1:
    {
	// Store Word From Coprocessor 1 /////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // swc1", pc, instr);
	return decode_swc1(instr);
    }
    case SWC2:
    {
	// Store Word From Coprocessor 2 /////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // swc2", pc, instr);
	process_reserved_instruction();
	return nothing_special;
    }
    case SCD:
    {
	// Store Conditional /////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tscd\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 2, 0))
	    process_address_error(data_store, va);
	PA pa = translate_vaddr(va, data_store);
	if (ll_bit)
	    store<doubleword>(gpr[rt(instr)], va, pa);
	gpr[rt(instr)] = ll_bit;
	sync_bit = true;
	ll_bit = false;
	return nothing_special;
    }
    case SDC1:
    {
	// Store Doubleword From Coprocessor 1 ///////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // sdc1", pc, instr);
	if (!allow_xinstr())
	    process_reserved_instruction();
	return decode_sdc1(instr);
    }
    case SDC2:
    {
	// Store Doubleword From Coprocessor 2 ///////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x // sdc2", pc, instr);
	process_reserved_instruction();
	return nothing_special;
    }
    case SD:
    {
	// Store Doubleword //////////////////////////////////////////////////
	if (trace_level >= print_instructions)
	    log("[%lx]\tsd\t%s, %d(%s)", pc,
		regname[rt(instr)],
		sign_extend<Int32>(offset(instr), 16),
		regname[base(instr)]);
	if (!allow_xinstr())
	    process_reserved_instruction();
	if (sync_bit)
	    sync();
	VA va = sign_extend<VA>(offset(instr), 16) + gpr[base(instr)];
	if (bits(va, 2, 0))
	    process_address_error(data_store, va);
	PA pa = translate_vaddr(va, data_store);
	store<doubleword>(gpr[rt(instr)], va, pa);
	return nothing_special;
    }
    default:
	// Reserved instruction.
	if (trace_level >= print_instructions)
	    log("[%lx]\t.long\t0x%08x", pc, instr);
	process_reserved_instruction();
	return nothing_special;
    }
}
