#ifndef koala_hh_included
#define koala_hh_included

#include <setjmp.h>
#include <mips64/bus.hh>
#include <mips64/cache.hh>
#include <mips64/cpu.hh>
#include <mips64/types.hh>
#include <checkpoint.hh>
#include <clock.hh>
#include <cpu.hh>
#include <inttypes.hh>
#include <serial.hh>

class Koala
    : public virtual Serializable, public MIPS64Cpu
{
    
private:

    static SerialType<Koala> type;

public:

    // The usual interfaces.
    Koala(const SimArgs& args);
    Koala(Checkpoint& cp);
    ~Koala();
    void reset(bool warm);
    void checkpoint(Checkpoint& cp, bool parent = false) const;

    // The fetch-execute loop.
    void run(ClockValue timeslice);

    // Interrupt delivery functions:
    void deliver_cold_reset();
    void deliver_soft_reset();
    void deliver_nmi();
    void deliver_interrupt(unsigned int n);
    void clear_interrupt(unsigned int n);
    void dump_state();

public:

    // Register names, for debugging.
    static const char* regname[32];

    // 16 bit implementation and revision fields.  The upper 8 bits are 0x20
    // for R4600, 0x21 for R4700. The lower 8 bits should distinguish the
    // simulator.

    static const UInt32 ImpRev = 0x2070;

    // A system coprocessor timer interrupt.
    class Timer
	: public Event, public virtual Serializable
    {
    private:
	static SerialType<Timer> type;
	Koala* cpu;
    public:
	Timer(Koala* c, ClockValue w)
	    : Event(w, 0), cpu(c) { }
	explicit Timer(const SimArgs& args)
	    : Event(args) { assert(UNREACHABLE); }
	explicit Timer(Checkpoint& cp);
	void invoke();
	void checkpoint(Checkpoint& cp, bool parent = false) const;
    };

    // Fundamental parameters.

    static const int vaddr_width 	= 40;
    static const int log2_icache_assoc 	=  1;
    static const int log2_icache_size  	= 14;
    static const int log2_icache_line  	=  5;
    static const int log2_dcache_assoc 	=  1;
    static const int log2_dcache_size  	= 14;
    static const int log2_dcache_line  	=  5;
    static const size_t write_buffers  	=  0;
    static const size_t tlb_size       	= 48;
    static const VA  xkseg_end		= VA(C000,00FF,8000,0000);

    // Some instruction latencies. These are for R4600. On R4700, both all
    // four multiplication latencies are decreased by four cycles.

    static const int mult_latency	= 10;
    static const int multu_latency	= 10;
    static const int div_latency	= 42;
    static const int divu_latency	= 42;
    static const int dmult_latency	= 12;
    static const int dmultu_latency	= 12;
    static const int ddiv_latency	= 74;
    static const int ddivu_latency	= 74;

    // Asynchronous event constants. Each interrupt source corresponds to a
    // single bit of the (events) word. Bits 8 through 16 correspond to
    // interrupts 0 through 7. Other bits represent the three reset exceptions
    // as follows:

    enum {
	cold_reset_event	= 1 << 0,
	soft_reset_event	= 1 << 1,
	nmi_event		= 1 << 2
    };

    // Cache of some strategic CPU mode bits. This shortcut is used to avoid
    // parsing of the somewhat convoluted Status register during address
    // translation and instruction decoding.
    enum {
	xmode = 1 << 0,	// 64 bit mode ([USK]X & [usk]mode)
	cmode = 1 << 1, // 32 bit (compatibility) mode (!xmode)
	bmode = 1 << 2,	// big-endian-cpu (big_endian_mem() ^ reverse_endian())
	rmode = 1 << 3, // reverse-endian (RE && umode)
	umode = 1 << 4,	// user mode
	smode = 1 << 5,	// supervisor mode
	kmode = 1 << 6	// kernel mode
    };

    // Pipeline information. This is fairly ad hoc, but it is still useful.

    enum {
	nothing_special,	// nothing special
	branch_delay,		// current instruction in the branch-delay slot
	instr_addr_error 	// instruction address error
    };

    // A TLB Entry structure.

    struct TLBEntry {
	TLBEntry* next;	// the hash chain
	TLBEntry* prev;	// the hash chain
	UInt64 hi;	// VPN/2 << 13
	UInt64 mask;	// ~((oddbit << 1) - 1)
	UInt64 oddbit;	// the odd bit of VPN
	UInt32 lo[2];	// EntryLo0/1
	Int16 asid;	// the ASID, sign bit set for global entries.
	UInt16 index;	// TLB index
	UInt16 hash;	// TLB map index
    };

    // An invalid ASID bit used to mark unused TLB entries.

    static const Int16 invalid_asid = (1 << 8);

    // An global ASID bit.

    static const Int16 global_asid = (Int16)(1 << 15);

    // Geometry of the TLB lookup map (see below.)

    static const int log2_tlb_map_size = 8;
    static const int tlb_map_size = 1 << log2_tlb_map_size;
 
    // Cache structures. See <mips64/cache.hh> for details.

    typedef MIPS64Cache<log2_icache_size,log2_icache_line,log2_icache_assoc>
	ICache;
    typedef MIPS64Cache<log2_dcache_size,log2_dcache_line,log2_dcache_assoc>
	DCache;
 
    // An invalid cache tag used to mark unused cache lines.

    static const UInt32 bad_tag = ~UInt32(0);

    // An invalid ibuf tag used to marked unused icache buffers.

    static const VA bad_ibuf_tag = ~(VA)(0);

    // An invalid physical address returned by translate_vaddr() on cache
    // operations that specify an invalid address.

    static const PA bad_pa = ~PA(0);

    // TLB access types, used to select the appropriate TLB miss handler.

    enum {
	instr_fetch,	// instruction fetch
	data_load,	// data load
	data_store,	// data store
	cache_op	// cache operations (ignore errors)
    };

    // The two-entry ITLB. Each entry maps a 4KB page.

    struct ITLBEntry {
	VA vpn;		// virtual address page number (va / 4KB)
	PA pa;		// physical address and the caching algorithm
	Int16 asid;	// ASID and the global bit.
    };

    // Although not in real hardware, we also cache the two most recent
    // I-cache accesses, as simulating cache lookups is slow, and as many as 8
    // instructions can be fetched from one cache line.

    struct ICacheBuffer {
	VA tag;		// address of the cache shifted by log2_icache_line
	UInt64* line;	// pointer to the ICache line.
	// Extract a doubleword given a PC.
	UInt64 operator[] (VA pc) const
	    { return line[bits(pc, Koala::log2_icache_line - 1, 3)]; }
    };

    // CPU state.

    Int16 asid;		// same as the ASID field in EntryHi
    UInt16 events;	// external events: this also replaces Cause[15:8].

    UInt8  pipeline;	// current instruction in the branch-delay slot
    UInt8  mode;	// CPU mode
    bool   sync_bit;	// true after executing the SYNC instruction.
    VA branch_target;	// next PC when in the branch delay slot

    // cp0[Random] is not updated on each clock cycle. Instead, its value is
    // updated when either cp0[Random] or cp0[Wired] is set, and the current
    // value is computed using the time elapsed since then.

    ClockValue random_seed;

    // Similary, we lazily compute the value of the Count register.

    ClockValue count_seed;

    // ITLB data. (lru_itlb) stores the index of the least-recently used entry.

    ITLBEntry itlb[2];
    bool lru_itlb;

    // I-cache buffer data (lru_ibuf) stores the index of the least-recently
    // used entry (0 or 1).

    ICacheBuffer ibuf[2];
    bool lru_ibuf;

    // There's at most one timer interrupt pending at any given time.
    
    Timer* timer;

    // The longjmp buffer used for handling exceptions.

    jmp_buf env;

    // The TLB and L1 caches.

    TLBEntry tlb[tlb_size];
    ICache icache;
    DCache dcache;

    // The TLB map used to simulate the direct-mapped TLB lookup.  There's an
    // extra entry at the end that contains the hash chain of all unused TLB
    // entries (that way, there is always exactly (tlb_size) mappings in the
    // hash table).

    TLBEntry* tlb_map[tlb_map_size + 1];

    // The trace mode flag.
    enum {
	no_tracing,
	report_interrupts,
	report_exceptions,
	print_instructions,
	dump_gprs
    };

    Int32 trace_level;

    // The NOP guard.
    static const UInt32 catch_nops = 128;
    size_t nop_count;

    // (mode) operations.

    void enter_kernel_mode()
    {
	mode = kmode | (bit(cp0[SR], SR_KX) ? xmode : cmode);
	if (big_endian_mem())
	    mode |= bmode;
    }

    void leave_kernel_mode()
    {
	switch (bits(cp0[SR], SR_KSU_Last, SR_EXL) << 1) {
	case 1 << SR_KSU_Last: // user
	    mode = umode | (bit(cp0[SR], SR_UX) ? xmode : cmode);
	    if (bit(cp0[SR], SR_RE))
		mode |= rmode;
	    if (big_endian_mem() != bool(reverse_endian()))
		mode |= bmode;
	    break;
	case 1 << SR_KSU_First: // supervisor
	    mode = smode | (bit(cp0[SR], SR_SX) ? xmode : cmode);
	    if (big_endian_mem())
		mode |= bmode;
	    break;
	default: // kernel (or undefined, but I ignore that bit ;)
	    mode = kmode | (bit(cp0[SR], SR_KX) ? xmode : cmode);
	    if (big_endian_mem())
		mode |= bmode;
	    break;
	}
    }

    // Compute the current value of the Random register.

    int get_random() const
    {
	int wired = cp0[Wired];
	int free = tlb_size - wired;
	return (tlb_size - 1 - wired + (now - random_seed)) % free + wired;
    }

    // An ASID is represented by a 16 bit integer that is either a positive
    // (non-zero) ASID, complemented  if the G bit is also set. asid_match()
    // checks two ASIDs for equivalence. Yet another demonstration of the
    // versatility of exclusive-or. ;)

    static bool asid_match(Int16 asid1, Int16 asid2)
	{ return (asid1 ^ asid2) <= 0; }

    // Operations on the TLB lookup map. The map is a hash table indexed by a
    // hash value computed from the ASID and the bits of virtual page number
    // unaffected by any page mask. The result of the map lookup is a TLB
    // entry. The hash function doesn't provide an ideal distribution, but as
    // a minimum, it maintains a separate hash chain per ASID, and provides
    // good distribution at least for sparse address spaces.  The hash table
    // always has an extra entry pointing to an invalid page.

    static int tlb_hash(VA va)
	{ return (bits(va, 63, 25) % tlb_map_size); }

    static bool va_match(VA va1, VA va2, VA mask)
	{ return !((va1 ^ va2) & mask); }

    // I-cache buffer operations.

    static bool ibuf_match(VA va, VA tag)
	{ return ((va >> log2_icache_line) ^ tag) == 0; }

    // TLB operations.

    void reset_tlb();
    TLBEntry* probe_tlb(VA va);
    void set_tlb_entry(int index);
    PA translate_vaddr(VA va, int type);

    // Instruction cache operations.

    void reset_icache();
    void control_icache(VA va, PA pa, int op, int type);
    Instr fetch(VA va, PA pa);

    // Data cache operations.

    void reset_dcache();
    void control_dcache(VA va, PA pa, int op, int type);
    template <int syscmd> UInt64 load(VA va, PA pa);		// data load
    template <int syscmd> void store(UInt64 x, VA va, PA pa);	// data store
    UInt64 load_left(UInt64 x, VA va, int syscmd);		// LDL, LWL
    UInt64 load_right(UInt64 x, VA va, int syscmd);		// LDR, LWR
    void store_left(UInt64 x, VA va, int syscmd);		// SDL, SWL
    void store_right(UInt64 x, VA va, int syscmd);		// SDR, SWR

    // Complete any pending memory operations.

    void sync();

    // Some state information constants.

    bool allow_xinstr() const
	{ return mode & (kmode|xmode); }
    bool branch_delay_slot() const
	{ return pipeline == branch_delay; }

    // Exception operations. process_exception() sets the simulator state up
    // for the exception handler as specified by 'cause' and 'vec'.
    // process_reset() processes a reset exception as specified by the
    // (events) word. All other functions handle particular exceptions by
    // calling process_exception() after setting any exception-specific
    // parameters.  Most of them never return as they longjmp to the start of
    // the pipeline.

    void process_reset();
    void process_exception(UInt32 cause, int vec);

    void process_address_error(int type, VA va)
    {
	int exc = (type == data_store) ? EXC_AdES : EXC_AdEL;
	cp0[BadVAddr] = va;
	process_exception(exc, common_vector);
    }

    void process_tlb_refill(int type, VA va, bool x)
    {
	int exc = (type == data_store) ? EXC_TLBS : EXC_TLBL;
	int vec = (bit(cp0[SR], SR_EXL)) ? common_vector :
	    (x) ? xtlb_refill_vector : tlb_refill_vector;
	cp0[BadVAddr] = va;
	cp0[Context] = set_bits(cp0[Context],
				bits(cp0[BadVAddr], 31, 13), 22, 4);
	VA badvpn2 = (bits(va, 63, 62) << 31) | bits(va, 39, 13);
	cp0[XContext] = set_bits(cp0[XContext], badvpn2, 32, 4);
	cp0[EntryHi] = (va & (bitmask(63,62) | bitmask(39, 13)))
	    | bits(asid, 7, 0);
	process_exception(exc, vec);
    }

    void process_tlb_invalid(int type, VA va)
    {
	int exc = (type == data_store) ? EXC_TLBS : EXC_TLBL;
	cp0[BadVAddr] = va;
	cp0[Context] = set_bits(cp0[Context],
				bits(cp0[BadVAddr], 31, 13), 22, 4);
	VA badvpn2 = (bits(va, 63, 62) << 31) | bits(va, 39, 13);
	cp0[XContext] = set_bits(cp0[XContext], badvpn2, 32, 4);
	cp0[EntryHi] = (va & (bitmask(63,62) | bitmask(39, 13)))
	    | bits(asid, 7, 0);
	process_exception(exc, common_vector);
    }

    void process_tlb_modified(VA va)
    {
	cp0[BadVAddr] = va;
	cp0[Context] = set_bits(cp0[Context],
				bits(cp0[BadVAddr], 31, 13), 22, 4);
	VA badvpn2 = (bits(va, 63, 62) << 31) | bits(va, 39, 13);
	cp0[XContext] = set_bits(cp0[XContext], badvpn2, 32, 4);
	cp0[EntryHi] = (va & (bitmask(63,62) | bitmask(39, 13)))
	    | bits(asid, 7, 0);
	process_exception(EXC_Mod, common_vector);
    }

    void process_bus_error(int type)
    {
	int exc = (type == instr_fetch) ? EXC_IBE : EXC_DBE;
	process_exception(exc, common_vector);
    }

    void process_integer_overflow()
	{ process_exception(EXC_Ov, common_vector); }
    void process_trap()
	{ process_exception(EXC_Tr, common_vector); }
    void process_syscall()
	{ process_exception(EXC_Sys, common_vector); }
    void process_breakpoint()
	{ process_exception(EXC_Bp, common_vector); }
    void process_reserved_instruction()
	{ process_exception(EXC_RI, common_vector); }
    void process_coprocessor_unusable(int c)
	{ process_exception(EXC_CpU | (c << Cause_CE_First), common_vector); }
    void process_fp_exception()
	{ process_exception(EXC_FPE, common_vector); }
    void process_interrupt()
	{ process_exception(EXC_Int, common_vector); }

    // Endianess parameters. The data received from the memory bus is always
    // in the host byte order, and uses host byte addressing.

    bool big_endian_mem() const
	{ return bit(cp0[Config], Config_BE); }
    bool reverse_endian() const
	{ return mode & rmode; }
    bool big_endian_cpu() const
	{ return mode & bmode; }

    // Helper used to extract data from a memory cell, taking into account
    // current endianess.

    template <int syscmd> typename FixedWidth<(syscmd + 1) * 8>::Unsigned
    swizzle(UInt64 x, unsigned addr) const
    {
	if (syscmd == word) {
	    // Optimization for instruction fetches.
	    UInt32 y = (((big_endian_cpu() << 2) ^ addr) & 7) ? x >> 32 : x;
	    return reverse_endian() ? byte_swap(y) : y;
	}
	else if (syscmd == doubleword) {
	    // Optimization for doublewords, assuming (addr & 7) == 0.
	    return reverse_endian() ? byte_swap(x) : x;
	}
	else {
	    // Generic implementation for all other types.
	    typedef typename FixedWidth<(syscmd + 1) * 8>::Unsigned T;
	    int shift = big_endian_cpu() ?
		(64 - IntTraits<T>::width) - (addr & 7) * 8 : (addr & 7) * 8;
	    T y = x >> shift;
	    return reverse_endian() ? byte_swap(y) : y;
	}
    }

    // Helper used to fix an address for a given type.

    template <int syscmd> PA swizzle(PA pa) const
    {
#if 0
	const PA mask = (syscmd == byte     ? bitmask(2, 0) :
			 syscmd == halfword ? bitmask(2, 1) :
			 syscmd == word     ? bitmask(2, 2) : 0);
	return (big_endian_cpu() == big_endian_host()) ? pa : pa ^ mask;
#endif
	return pa;
    }

    // Set the Coprocessor 0 timer interrupt.

    void set_timer();

    // Coprocessor 0 register access. This is necessary as some registers
    // are computed on demand, and most registers have some read-only fields.

    UInt64 read_cp0(int n) const;
    void write_cp0(int n, UInt64 x);

    // Coprocessor 1 (FPU) simulator:

    // Floating point condition names
    static const char* condname[16];

    enum {
	FCR0  =  0,
	FP_Rev_Last = 7, FP_Rev_First = 0,
	FP_Imp_Last = 15, FP_Imp_First = 8,
	FCR31 = 31,
	FP_RM_Last = 1, FP_RM_First = 0,
	FP_Flag_Last = 6, FP_Flag_First = 2,
	FP_Enable_Last = 11, FP_Enable_First = 7,
	FP_Cause_Last = 17, FP_Cause_First = 12,
	FP_C = 23,
	FP_FS = 24
    };

    // Software IEC/IEEE floating-point rounding mode.
    enum {
	float_round_nearest_even	= 0,
	float_round_to_zero      	= 1,
	float_round_up           	= 2,
	float_round_down         	= 3
    };
    
    // Software IEC/IEEE floating-point exception flags.
    enum {
	float_flag_inexact   		=  1,
	float_flag_underflow 		=  2,
	float_flag_overflow  		=  4,
	float_flag_divbyzero 		=  8,
	float_flag_invalid   		= 16,
	float_flag_unimplemented	= 32
    };

    int float_rounding_mode() const
	{ return bits(cp1[FCR31], FP_RM_Last, FP_RM_First); }

    float convert_from_s(UInt32 value) const
	{ return *(float *)(&value); }

    double convert_from_d(UInt64 value)
	{ return *(double *)(&value); }

    UInt32 convert_to_s(float value)
	{ return *(UInt32 *)(&value); }

    UInt64 convert_to_d(double value)
	{ return *(UInt64 *)(&value); }

    int decode_cop1_s(Instr instr);
    int decode_cop1_d(Instr instr);
    int decode_cop1_w(Instr instr);
    int decode_cop1_l(Instr instr);

    // The main instruction decoder with satelite functions for some of the
    // more elaborate instructions.

    int decode(Instr instr);
    int decode_cop0(Instr instr);
    int decode_cop1(Instr instr);
    int decode_cache(Instr instr);
    int decode_ldc1(Instr instr);
    int decode_lwc1(Instr instr);
    int decode_sdc1(Instr instr);
    int decode_swc1(Instr instr);

    // Some debugging help.
    void dump_gpr_registers() const;
    void dump_fpr_registers() const;

private:

    // Finally, some configuration data.
    struct {
	char* bus;
	int ec;
	char* ep;
	int be;
	int trace;
    } conf;

};


#endif // koala_hh_included

